// Copyright 2011 Julian Phillips.  All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package py

// #include "utils.h"
//import "C"

import (
	"sync"
	"unsafe"
)

type CFunction struct {
	AbstractObject
	o _PyCFunctionObject
}

func cfunctionCheck(obj Object) bool {
	if obj == nil {
		return false
	}
	return _cFunctionCheck(c(obj)) != 0
}

func newCFunction(obj *_PyObject) *CFunction {
	return (*CFunction)(unsafe.Pointer(obj))
}

func NewCFunction(name string, fn interface{}, doc string) (*CFunction, error) {
	return makeCFunction(name, fn, doc, nil)
}

func makeCFunction(name string, fn interface{}, doc string, mod_name *_PyObject) (*CFunction, error) {
	ml := _newMethodDef()

	switch fn.(type) {

	case func() (Object, error):
		_set_call_noargs(&ml.ml_meth)
		ml.ml_flags = _METH_NOARGS

	case func(a *Tuple) (Object, error):
		_set_call_args(&ml.ml_meth)
		ml.ml_flags = _METH_VARARGS

	case func(a *Tuple, k *Dict) (Object, error):
		_set_call_keywords(&ml.ml_meth)
		ml.ml_flags = _METH_VARARGS | _METH_KEYWORDS

	default:
		_free(unsafe.Pointer(ml))
		return nil, TypeError.Err("CFunction_New: unknown func type for %s", name)

	}

	ret := _PyCFunction_NewEx(ml, saveFunc(fn), mod_name)
	if ret == nil {
		_free(unsafe.Pointer(ml))
		return nil, exception()
	}

	ml.ml_name = _CString(name)
	ml.ml_doc = _CString(doc)

	return newCFunction(ret), nil
}

// PyCFunction_GetFunction

func (f *CFunction) Self() (Object, error) {
	ret := _PyCFunction_GetSelf(c(f))
	if ret == nil {
		return nil, exception()
	}
	return newObject(ret), nil
}

func (f *CFunction) Flags() (int, error) {
	ret := _PyCFunction_GetFlags(c(f))
	return int(ret), exception()
}

func (f *CFunction) Call(args *Tuple, kw *Dict) (Object, error) {
	ret := _PyCFunction_Call(c(f), c(args), c(kw))
	if ret == nil {
		return nil, exception()
	}
	return newObject(ret), nil
}

type Method struct {
	Name string
	Func interface{}
	Doc  string
}

var (
	funcLock sync.RWMutex
	funcs    []interface{}
)

func saveFunc(f interface{}) *_PyObject {
	funcLock.Lock()
	defer funcLock.Unlock()

	funcs = append(funcs, f)
	return _PyInt_FromLong(int64(len(funcs) - 1))
}

func getFunc(self unsafe.Pointer) interface{} {
	funcLock.RLock()
	defer funcLock.RUnlock()

	idx := int(_PyInt_AsLong((*_PyObject)(self)))

	if idx >= len(funcs) {
		return nil
	}

	return funcs[idx]
}

//export CallWithoutArgs
func CallWithoutArgs(self, args unsafe.Pointer) unsafe.Pointer {
	f, ok := getFunc(self).(func() (Object, error))
	if !ok {
		raise(AssertionError.Err("CallWithoutArgs: wrong function type!!!"))
		return nil
	}
	ret, err := f()
	if err != nil {
		raise(err)
		return nil
	}
	return unsafe.Pointer(c(ret))
}

//export CallWithArgs
func CallWithArgs(self, args unsafe.Pointer) unsafe.Pointer {
	f, ok := getFunc(self).(func(a *Tuple) (Object, error))
	if !ok {
		raise(AssertionError.Err("CallWithArgs: wrong function type!!!"))
		return nil
	}
	a := newTuple((*_PyObject)(args))
	ret, err := f(a)
	if err != nil {
		raise(err)
		return nil
	}
	return unsafe.Pointer(c(ret))
}

//export CallWithKeywords
func CallWithKeywords(self, args, kw unsafe.Pointer) unsafe.Pointer {
	f, ok := getFunc(self).(func(a *Tuple, k *Dict) (Object, error))
	if !ok {
		raise(AssertionError.Err("CallWithKeywords: wrong function type!!!"))
		return nil
	}
	a := newTuple((*_PyObject)(args))
	k := newDict((*_PyObject)(kw))
	ret, err := f(a, k)
	if err != nil {
		raise(err)
		return nil
	}
	return unsafe.Pointer(c(ret))
}
