// Copyright 2011 Julian Phillips.  All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package py

// #include "utils.h"
//import "C"

import (
	"reflect"
	"sync"
	"unsafe"
)

type Op int

const (
	LT = Op(_Py_LT)
	LE = Op(_Py_LE)
	EQ = Op(_Py_EQ)
	NE = Op(_Py_NE)
	GT = Op(_Py_GT)
	GE = Op(_Py_GE)
)

// Object is the generic interface that represents a Python object.  All of the
// concrete types satisfy the Object interface.
type Object interface {
	Base() *BaseObject
	Type() *Type
	Decref()
	Incref()
	IsTrue() bool
	Not() bool
	Free()
}

// None is the Python equivalent to nil.
var None = (*NoneObject)(unsafe.Pointer(&*__Py_NoneStruct))

// NoneObject is the type of the None value.  The only value of this type is
// None.
type NoneObject struct {
	AbstractObject
}

func (n *NoneObject) String() string {
	return "None"
}

func c(obj Object) *_PyObject {
	if obj == nil {
		return nil
	}
	return (*_PyObject)(unsafe.Pointer(obj.Base()))
}

var (
	typeLock sync.RWMutex
	types    = make(map[*_PyTypeObject]*Class)
)

func registerType(pyType *_PyTypeObject, class *Class) {
	typeLock.Lock()
	defer typeLock.Unlock()

	types[pyType] = class
}

func getType(pyType *_PyTypeObject) (*Class, bool) {
	typeLock.RLock()
	defer typeLock.RUnlock()

	class, ok := types[pyType]

	return class, ok
}

func obj2Class(c *Class, obj *_PyObject) (Object, bool) {
	vp := reflect.NewAt(reflect.TypeOf(c.Pointer), unsafe.Pointer(&obj))
	o, ok := vp.Elem().Interface().(Object)
	return o, ok
}

func newObject(obj *_PyObject) Object {
	if obj == nil {
		return nil
	}

	o := unsafe.Pointer(obj)
	if o == unsafe.Pointer(None) {
		return None
	}

	pyType := (*_PyTypeObject)(obj.ob_type)
	class, ok := getType(pyType)
	if ok {
		ret, ok := obj2Class(class, obj)
		if ok {
			return ret
		}
	}
	for pyType.tp_base != nil {
		pyType = (*_PyTypeObject)(unsafe.Pointer(pyType.tp_base))
		class, ok := getType(pyType)
		if ok {
			ret, ok := obj2Class(class, obj)
			if ok {
				return ret
			}
		}
	}

	switch _getBasePyType(obj) {
	case &*_PyList_Type:
		return (*List)(o)
	case &*_PyTuple_Type:
		return (*Tuple)(o)
	case &*_PyDict_Type:
		return (*Dict)(o)
	case &*_PyString_Type:
		return (*String)(o)
	case &*_PyBool_Type:
		return newBool(obj)
	case &*_PyInt_Type:
		return (*Int)(o)
	case &*_PyLong_Type:
		return (*Long)(o)
	case &*_PyFloat_Type:
		return (*Float)(o)
	case &*_PyModule_Type:
		return (*Module)(o)
	case &*_PyType_Type:
		return (*Type)(o)
	case &*_PyCode_Type:
		return (*Code)(o)
	case &*_PyCFunction_Type:
		return (*CFunction)(o)
	case &*_PyComplex_Type:
		return (*Complex)(o)
	case &*_PyFrozenSet_Type:
		return (*FrozenSet)(o)
	case &*_PySet_Type:
		return (*Set)(o)
	case &*_PyFunction_Type:
		return (*Function)(o)
	}

	if _exceptionCheck(obj) != 0 {
		return newException(obj)
	}

	return newBaseObject(obj)
}
